/*
  Copyright 1996-2000
  Simon Whiteside

* $Id: ExecutableRoot.java,v 1.14 2001/02/10 15:01:25 sdw Exp $
*/

package simkin;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
/**
 * This class provides an implementation of the Executable interface which can 
 * be conveniently extended
 * <p>The class implements some commonly used methods via the {@link simkin.ExecutableRoot#method method} function:<ul>
 * <li>the method "trace" which outputs a message to a tracer.
 * <li>the method "isObject" which returns true if the argument is an object
 * <li>the method "length" which returns the length of a string
 * <li>the method "charAt" which returns a character within a string </ul>
 * <p>It also implements all the methods of the {@link simkin.Executable Executable} interface which means that derived classes need only implement the methods they need to override.
 * <p>The class uses the Interpreter's reflection methods to look for public fields and method names which match those being accessed by the Simkin script. 
 */
public class ExecutableRoot implements Executable {
    /** 
     * This variable stores a callback which is called when subclasses call the Trace method
     */
    static TraceCallback g_Tracer;
    /**
     * this is the name of the trace method which this class exposes
     */
    static final String s_trace="trace";
    /**
     * this is the name of the isObject method which this class exposes
     */
    static final String s_isObject="isObject";
    /**
     * this is the name of the length method which this class exposes
     */
    static final String s_length="length";
    /**
     * this is the name of the charAt method which this class exposes
     */
    static final String s_charAt="charAt";
    /**
     * This method sets the object which will receive trace calls when a trace is requested by a subclass
     * @param cb the object which will be called to output a trace message
     */
    public static void setTracer(TraceCallback cb){
	g_Tracer=cb;
    }
    /**
     * This method calls the Interpreter's reflectiveSetValue method
     */
    public void setValue(String field_name,String attrib,Object value)  throws RuntimeException,NoSuchFieldException{
	Interpreter.getInterpreter().reflectiveSetValue(this,field_name,value);
    }
    /**
     * This method calls the Interpreter's reflectiveGetValue method
     */
    public Object getValue(String field_name,String attrib) throws RuntimeException,NoSuchFieldException{
	return Interpreter.getInterpreter().reflectiveGetValue(this,field_name);
    }
    /**
     * The following methods are supported directly by this class:
     * <p> 
     * <p>trace(a) - sends a message to the trace output, takes a variable number of arguments
     * <p>isObject(a) - returns true if a is an object (as opposed to an integer or string)
     * <p>length(a) - returns the length of a (taken as a string)
     * <p>charAt(str,index) - returns the character at the given index within the string
     * <p>If the method is not one of these, the Interpreter's reflectiveMethodCall method is called
     * @return true if the method is found
     */
    public Object method(String method_name,Object[] arguments) throws ParseException, RuntimeException,NoSuchMethodException{
	Object ret=null;
	if (method_name.equals(s_trace) && arguments!=null){
	    StringBuffer buffer=new StringBuffer();
	    for (int i=0;i<arguments.length;i++)
		buffer.append(arguments[i].toString());
	    if (g_Tracer!=null)
		g_Tracer.trace(buffer.toString());
	    else
		Tracer.trace(buffer.toString());
	}else if (method_name.equals(s_isObject) && arguments!=null){
		Object arg=arguments[0];
		if ((arg instanceof Integer || arg instanceof Double || arg instanceof Boolean || arg instanceof Character || arg instanceof String))
		    ret=Boolean.FALSE;
		else
		    ret=Boolean.TRUE;
	}else if (method_name.equals(s_length) && arguments!=null)
	    ret=new Integer(arguments[0].toString().length());
	else if (method_name.equals(s_charAt) && arguments!=null && arguments.length>1)
	    ret=new Character(arguments[0].toString().charAt(Interpreter.intValue(arguments[1])));
	else
	    ret=Interpreter.getInterpreter().reflectiveMethodCall(this,method_name,arguments);
	return ret;
    }
};
