/*
  Copyright 1996-2000
  Simon Whiteside

  $Id: Interpreter.java,v 1.19 2001/02/10 15:01:25 sdw Exp $
*/
package simkin;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.lang.reflect.Method;
import java.lang.reflect.Field;

/**
 * This class represents the Simkin interpreter. The interpreter can load, parse and interpret Simkin code.
 * <p>The Interpreter maintains a list of global variables. This maps names to objects, which are available to all Simkin scripts.
 * <p>The interpreter itself is a global object with the name "Interpreter".
 * <p>The Interpreter implements the {@link simkin.Executable Executable} interface, supports the single field "tracing" which enables/disables tracing of method calls
 */
public class Interpreter implements Executable {
    /**
     * this is the interpreter used by objects when a call into simkin is required
     */
    static Interpreter g_Interpreter;
    /**
     * this string holds the location of the current method
     */
    String m_Location;
    /**
     * this is the name of the interpreter global variable
     */
    static String s_Interpreter="Interpreter";
    /**
     * this is the name of the Interpreter's tracing variable, if true, the interpreter prints out tracing information
     */
    static String s_Tracing="tracing";
    /**
     * This table holds the global variables for this interpreter
     */
    Hashtable m_GlobalVars=new Hashtable();
    /**
     * This constant shows the number of local variable tables that are cached
     */
    static final int MAX_LOCAL_VARS_CACHE=5;
    /**
     * this is a pool of local variable tables to speed up method calling
     */
    Hashtable	m_LocalVars[];
    /**
     * this variable shows how deeply nested the stack is in this interpreter
     */
    int m_StackDepth=0;
    /**
     * if this variable is true, the interpreter outputs information about the execution of the code (method calls and return values)
     */
    boolean m_Tracing=false;
    /**
     * this variable sets tracing on for the functionality of the interpreter
     */
    static final boolean g_TraceExecution=false;
    /**
     * the null object
     */
    static final Null g_null=new Null();    
    /**
     * Default constructor - adds a global variable pointing to this interpreter
     */
    public Interpreter(){
	addGlobalVariable(s_Interpreter,this);
	m_LocalVars=new Hashtable[MAX_LOCAL_VARS_CACHE];
	for (int i=0;i<MAX_LOCAL_VARS_CACHE;i++)
	    m_LocalVars[i]=new Hashtable();
    }
    /**
     * Used to set the "tracing" variable - for example in Simkin:
     * <p> Interpreter.tracing=true;
     */
    public void setValue(String field_name,String attrib,Object value) throws RuntimeException,NoSuchFieldException{
	if (field_name.equals(s_Tracing)){
	    m_Tracing=boolValue(value);
	}else
	    throw new NoSuchFieldException();
    }
    /**
     * The interpreter does not expose any methods to Simkin scripts
     */
    public Object method(String method_name,Object[] args) throws RuntimeException,ParseException,NoSuchMethodException{
	throw new NoSuchMethodException();
    }
    /**
     * This method converts the given value into a boolean value
     * <p>If the object is an instance of Boolean, the boolean value is returned, otherwise toString() is called, and
     * compared with "true"
     * @return the boolean equivalent of the object
     */
    public static boolean boolValue(Object value){
	boolean ret=false;
	if (value!=null){
	    if (value instanceof Boolean)
		ret=((Boolean)value).booleanValue();
	    else
		if (value.toString().equals(NodeTypes.s_True))
		    ret=true;
	}
	return ret;
    }
    /**
     * This method converts the given value into an integer value
     * <p>If the object is an instance of Integer, the integer value is returned, otherwise toString() is called, and
     * converted to an integer
     * @return the integer equivalent of the object
     */
    public static int intValue(Object value){
	int ret=0;
	if (value instanceof Integer)
	    ret=((Integer)value).intValue();
	else{
	    try{
		ret=Integer.parseInt(value.toString());
	    }catch(Exception e){
	    }
	}
	return ret;
    }
    /**
     * This method converts the given value into a double value
     * <p>If the object is an instance of Double, the integer value is returned, otherwise toString() is called, and
     * converted to a double
     * @return the double equivalent of the object
     */
    public static double floatValue(Object value){
	double ret=0;
	if (value instanceof Double)
	    ret=((Double)value).doubleValue();
	else{
	    try{
		ret=new Double(value.toString()).doubleValue();
	    }catch(Exception e){
	    }
	}
	return ret;
    }
    /**
     * Returns the value of the "tracing" variable - e.g. in Simkin:
     * <p> tracing=Interpreter.tracing;
     */
    public Object getValue(String field_name,String attrib) throws RuntimeException,NoSuchFieldException{
	Object o=null;
	boolean bRet=false;
	if (field_name.equals(s_Tracing))
	    o=new Boolean(m_Tracing);
	else
	    throw new NoSuchFieldException();
	return o;
    }
    /**
     * Called to execute a statement
     * @return a return object
     */
    Object executeStat(Object obj,Hashtable var,StatNode stat) throws ParseException, RuntimeException{
	if (g_TraceExecution) Tracer.trace("executeStat "+NodeTypes.g_NodeNames[stat.getType()]);
	Object ret=null;
	switch(stat.getType()){
	case NodeTypes.s_StatList:
	    ret=executeStats(obj,var,(StatListNode)stat);
	    break;
	case NodeTypes.s_Method:
	    evalMethod(obj,var,((MethodNode)stat).m_Ids);
	    break;
	case NodeTypes.s_if:
	    ret=executeIfStat(obj,var,(IfNode)stat);
	    break;
	case NodeTypes.s_while:
	    ret=executeWhileStat(obj,var,(WhileNode)stat);
	    break;
	case NodeTypes.s_Assign:
	    executeAssignStat(obj,var,(AssignNode)stat);
	    break;
	case NodeTypes.s_return:
	    ret=executeReturnStat(obj,var,(ReturnNode)stat);
	    break;
	case NodeTypes.s_switch:
	    ret=executeSwitchStat(obj,var,(SwitchNode)stat);
	    break;
	}
	return ret;
    }
    /**
     * Called to execute a block of statements
     * @return a return object
     */
    Object executeStats(Object obj,Hashtable var,StatListNode n) throws ParseException, RuntimeException{
	if (g_TraceExecution) Tracer.trace("executeStats "+n.m_Stats.size());
	Object ret=null;
	for (int i=0;i<n.m_Stats.size();i++){
	    StatNode stat=(StatNode)(n.m_Stats.elementAt(i));
	    ret=executeStat(obj,var,stat);
	    if (ret!=null)
		break;
	}
	return ret;
    }
    /**
     * Adds a local variable to the current list of local variables
     */
    void addLocalVariable(Hashtable var,String name,Object value){
	if (g_TraceExecution) Tracer.trace("Adding local variable:" +name);
	    var.put(name,value);
    }
    /**
     * looks for the given global variable.
     * @param name - the name of the global variable
     * @return the object found, or null if the global variable is undefined
     */
    public Object findGlobalVariable(String name){
	return m_GlobalVars.get(name);
    }
    /**
     * Adds a global variable to the current list
     * @param name - the name of the global variable, if one already exists it is replaced
     * @param r - the object to be added to the list
     */
    public void addGlobalVariable(String name,Object r){
	m_GlobalVars.put(name,r);
    }
    /**
     * Removes a global variable from the current list
     * @param name - the name of the global variable to be removed
     */
    public void removeGlobalVariable(String name){
	m_GlobalVars.remove(name);
    }
    /**
     * Evaluates a method and returns the result
     * @param obj the object owning the code
     * @param var the current local variable list
     * @param ids the id list
     * @return the value from the method call
     */
    Object evalMethod(Object obj,Hashtable var,IdListNode ids) throws ParseException, RuntimeException{
	Object ret=null;
	IdNode idNode=(IdNode)ids.m_Ids.elementAt(ids.m_Ids.size()-1);
	String method_name=idNode.m_Name;
	if (ids.m_Ids.size()==1){
	    ret=makeMethodCall(obj,var,method_name,idNode.m_Exprs);
	}else{
	    obj=followIdList(obj,var,ids);
	    ret=makeMethodCall(obj,var,method_name,idNode.m_Exprs);
	}
	return ret;
    }
    /**
     * executes a return statement by copying the expression's value into the return variable
     * @return the result of the result expression
     */
    Object executeReturnStat(Object obj,Hashtable var,ReturnNode n) throws ParseException, RuntimeException{
	return evaluate(obj,var,n.m_Expr);
    }
    /**
     * executes a while statement by repeatedly executing the block of statements until the test expression is false
     * @return the return object
     */
    Object executeWhileStat(Object obj,Hashtable var,WhileNode n) throws ParseException, RuntimeException{
	Object ret=null;
	// loop until an enclosing statement is a return statement
	while(ret==null){
	    Object expr=evaluate(obj,var,n.m_Expr);
	    // or until the loop expression is false
	    if (boolValue(expr)){
		if (n.m_Stats!=null)
		    ret=executeStat(obj,var,n.m_Stats);
	    }else
		break;
	}
	return ret;
    }
    /**
     * executes a switch statement by evaluating an expression and testing it for equality against the case statements
     * @return the return object
     */
    Object executeSwitchStat(Object obj,Hashtable var,SwitchNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object expr=evaluate(obj,var,n.m_Expr);
	CaseNode caseFound=null;
	for (int i=0;i<n.m_Cases.m_Cases.size();i++){
	    if (g_TraceExecution) Tracer.trace("checking case "+i);
	    CaseNode caseNode=(CaseNode)(n.m_Cases.m_Cases.elementAt(i));
	    Object testExpr=evaluate(obj,var,caseNode.m_Expr);
	    if ((expr==null && testExpr==null) || (expr!=null && expr.equals(testExpr))){
		caseFound=caseNode;
		break;
	    }
	}
	if (caseFound!=null){
	    ret=executeStats(obj,var,caseFound.m_Stats);
	}else if (n.m_DefaultCase!=null)
	    ret=executeStats(obj,var,n.m_DefaultCase.m_Stats);
	return ret;
    }
    /**
     * Executes an if..else statement
     * @return the return object
     */
    Object executeIfStat(Object obj,Hashtable var,IfNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object rExpr=evaluate(obj,var,n.m_Expr);
	if (boolValue(rExpr)){
	    if (n.m_Stats!=null)
		ret=executeStat(obj,var,n.m_Stats);
	}else
	    if (n.m_Else!=null)		
		ret=executeStat(obj,var,n.m_Else);
	return ret;
    }
    /**
     * Evaluates a series of argument expressions for a method call
     * @param obj - the object owning the script
     * @param var - the current local variables
     * @param n - the node containing the list of expressions
     * @param args - the array of arguments to be filled up
     */
    void evalArgs(Object obj,Hashtable var,ExprListNode n,Object [] args) throws ParseException, RuntimeException{
	if (n!=null && n.m_Exprs!=null)
	    for (int i=0;i<n.m_Exprs.size();i++){
		ExprNode parg=(ExprNode)(n.m_Exprs.elementAt(i));
		args[i]=evaluate(obj,var,parg);
	    }
    }
    /**
     * Checks whether an identifier is prefixed by "@". If it is, the id is first evaluated and then a variable with the value's name returns as the new id
     * @return the name, once the indirection has been resolved
     */
    String checkIndirectId(Object obj,Hashtable var,String name) throws ParseException{
	if (name.charAt(0)=='@'){
	    name=name.substring(1,name.length());
	    Object new_name=findValue(obj,var,name,null);
	    if (new_name!=null)
		name=new_name.toString();
	}
	return name;
    }	
    /**
     * searches for a named value. Looking first for built-in names, then within the given local variable table, then within the given object, and finally within the global variables
     * @param obj the object whose fields will be searched
     * @param var the local variable list
     * @param name the name of the value being sought
     * @param attrib the name of the attribute being sought (might be null)
     * @return the object found
     */
    Object findValue(Object obj,Hashtable var,String name,String attrib) throws ParseException{
	Object ret=null;
	name=checkIndirectId(obj,var,name);
	if (name.length()>0){
	    // the name is prefixed with "self"
	    if (name.equals(NodeTypes.s_null)){
		ret=g_null;
	    }else if (name.equals(NodeTypes.s_Self))
		ret=extractValue(obj,null,attrib);
	    // the value is "true"
	    else if (name.equals(NodeTypes.s_True))
	    ret=Boolean.TRUE;
		// the value is "false"
	    else if (name.equals(NodeTypes.s_False))
		ret=Boolean.FALSE;
	    else{
		// otherwise first check in the local variables
		ret=var.get(name);
		if (ret==null){
		    // then in the instance data for the object
		    try{
			if (g_TraceExecution) Tracer.trace("looking for instance var:"+name);
			if (obj!=null)
			    ret=extractValue(obj,name,attrib);
			else{
			    ret=extractValue(m_GlobalVars.get(name),null,attrib);
			    if (ret==null)
				runtimeException("Field "+name+" not found\n");
			}
		    }catch(Exception e){
			if (g_TraceExecution) Tracer.trace("looking for global var:"+name);
			// finally in the global variables list
			ret=extractValue(m_GlobalVars.get(name),null,attrib);
			if (ret==null)
			    runtimeException("Field "+name+" not found\n");
		    }
		}else
		    ret=extractValue(ret,null,attrib);
	    }
	}	
	return ret;
    }
    /**
     * This method extracts a value from an object
     * @param obj - the owner of the value
     * @param name - the name of the field (can be null)
     * @param attrib - the name of an attribute (can be null)
     */
    Object extractValue(Object obj,String name,String attrib) {
	if (g_TraceExecution) Tracer.trace("extractValue:"+name+" "+attrib);
	Object ret=obj;
	if (attrib!=null){
	    if (obj instanceof Executable){
		try{
		    ret=((Executable)obj).getValue(name,attrib);
		}catch(Exception e){
		    runtimeException("Attribute "+attrib+" not found\n");
		}
	    }else{
		runtimeException("Attributes not supported on non-Executable objects\n");
	    }
	}else{
	    if (name!=null){
		if (obj instanceof Executable){
		    try{
			ret=((Executable)obj).getValue(name,attrib);
		    }catch(Exception e){
			runtimeException("Field "+name+" not found\n");
		    }
		}else{
		    try{
			ret=reflectiveGetValue(obj,name);
		    }catch(Exception e){
			runtimeException("Field "+name+" not found\n");
		    }
		}
	    }
	}
	if (ret==null)
	    ret=g_null;
	return ret;
    }
    /**
     * this method assigns the value of an expression to a field in another object
     * @param obj the object in which context the method is executed
     * @param var the local variable list
     * @param n the portion of the parse tree for the assignment statement
     */
    void executeAssignStat(Object obj,Hashtable var,AssignNode n) throws ParseException, RuntimeException{
	Object value=evaluate(obj,var,n.m_Expr);
	IdNode idNode=(IdNode)(n.m_Ids.m_Ids.elementAt(n.m_Ids.m_Ids.size()-1));
	String name=checkIndirectId(obj,var,idNode.m_Name);
	if (n.m_Ids.m_Ids.size()==1){
	    if (name.equals(NodeTypes.s_Self)){
		try{
		    insertValue(obj,null,n.m_Attrib,value);
		}catch(Exception e){
		    runtimeException("Could not set attribute "+n.m_Attrib+"\n");
		}
	    }else{
		try{
		    insertValue(obj,name,n.m_Attrib,value);
		}catch(NoSuchFieldException e){
		    addLocalVariable(var,name,value);
		}catch(Exception e){
		    runtimeException("Could not set field "+name+" attribute "+n.m_Attrib+"\n");
		}
	    }
	}else{
	    obj=followIdList(obj,var,n.m_Ids);
	    try{
		insertValue(obj,name,n.m_Attrib,value);
	    }catch(Exception e){
		runtimeException("Could not set field "+name+" attribute "+n.m_Attrib+"\n");
	    }
	}
    }
    /**
     * This method sets a value into an object
     * @param obj - the object owning the value
     * @param name - the name of the field
     * @param attrib - the attribute name
     * @param value - the value to be set
     */
    void insertValue(Object obj,String name,String attrib,Object value) throws NoSuchFieldException{
	if (attrib!=null){
	    if (obj instanceof Executable){
		((Executable)obj).setValue(name,attrib,value);
	    }else{
		runtimeException("Attributes not supported on non-Executable objects\n");
	    }
	}else{
	    if (obj instanceof Executable)
		((Executable)obj).setValue(name,attrib,value);
	    else
		reflectiveSetValue(obj,name,value);
	}
    }
    /**
     * This method follows an id list to the penultimate id, attempting to
     * locate objects
     * @param obj - owner of this method
     * @param var - current local variables
     * @param idlist - list of id's
     * @return object - the object found by following the list
     */
    Object followIdList(Object obj,Hashtable var,IdListNode idList) throws ParseException{
	IdNode idNode=(IdNode)idList.m_Ids.elementAt(0);
	String name=checkIndirectId(obj,var,idNode.m_Name);
	if (idNode.m_Exprs==null)
	    obj=findValue(obj,var,name,null);
	else
	    obj=makeMethodCall(obj,var,name,idNode.m_Exprs);
	for (int i=1;i<idList.m_Ids.size()-1;i++){
	    idNode=(IdNode)idList.m_Ids.elementAt(i);
	    name=checkIndirectId(obj,var,idNode.m_Name);
	    if (idNode.m_Exprs==null)
		obj=extractValue(obj,name,null);
	    else
		obj=makeMethodCall(obj,var,name,idNode.m_Exprs);
	}
	return obj;
    }
    /**
     * This method calls a method on the given object
     * @param obj - the object whose method is to be called
     * @param var - the current local variables
     * @param name - the name of the method
     * @param exprs - the expression list to be passed as parameters - these will be evaluated
     */
    Object makeMethodCall(Object obj,Hashtable var,String name,ExprListNode exprs) throws ParseException{
	String method_name=checkIndirectId(obj,var,name);
	Object [] args=null;
	Object ret=null;
	if (exprs!=null && exprs.m_Exprs.size()>0)
	    args=new Object[exprs.m_Exprs.size()];
	evalArgs(obj,var,exprs,args);
	if (m_Tracing)  traceString("Interpreter: "+obj.getClass().getName()+"::"+method_name+"()\n");
	try{
	    if (obj instanceof Executable){
		ret=((Executable)obj).method(method_name,args);	
	    }else
		ret=reflectiveMethodCall(obj,method_name,args);
	}catch(NoSuchMethodException e){
	    runtimeException("Method "+method_name+" could not be called:"+e.toString()+"\n");
	}catch(Exception e){
	    Tracer.trace(e);
	    runtimeException(e.toString());
	}
	if (m_Tracing)
	    traceString("Interpreter: "+method_name+"() returns "+(ret!=null?ret.toString():"null")+"\n");
	return ret;
    }
    /** 
     * This method evaluates an expression
     * @param obj the object in whose context the expression is evaluated
     * @param var the local variable list
     * @param n the portion of the parse tree for this expression
     * @return the value of the expression
     */
    Object evaluate(Object obj,Hashtable var,ExprNode n) throws ParseException, RuntimeException{
	Object ret=null;
	Object other=null;
	if (g_TraceExecution) Tracer.trace("evaluate "+NodeTypes.g_NodeNames[n.getType()]);
	switch(n.getType()){
	case NodeTypes.s_IdList:{
	    IdListNode fnode=(IdListNode)n;
	    IdNode idNode=(IdNode)fnode.m_Ids.elementAt(fnode.m_Ids.size()-1);
	    String name=idNode.m_Name;
	    if (g_TraceExecution) Tracer.trace("evaluate "+NodeTypes.g_NodeNames[n.getType()]+" name \""+name+"\"");
	    if (fnode.m_Ids.size()==1){
		if (idNode.m_Exprs==null)
		    ret=findValue(obj,var,name,fnode.m_Attrib);
		else
		    ret=makeMethodCall(obj,var,name,idNode.m_Exprs);
	    }else{
		obj=followIdList(obj,var,fnode);
		if (idNode.m_Exprs==null)
		    ret=extractValue(obj,name,fnode.m_Attrib);
		else{
		    ret=makeMethodCall(obj,var,name,idNode.m_Exprs);
		    if (fnode.m_Attrib!=null)
			ret=extractValue(ret,null,fnode.m_Attrib);
		}
	    }
	    break;
	}
	case NodeTypes.s_String:
	    ret=((LiteralNode)n).m_Value;
	    break;
	case NodeTypes.s_Character:
	    ret=new Character(((LiteralNode)n).m_Value.charAt(0));
	    break;
	case NodeTypes.s_Integer:
	    ret=new Integer(((LiteralNode)n).m_Value);
	    break;
	case NodeTypes.s_Float:
	    ret=new Double(((LiteralNode)n).m_Value);
	    break;
	default:{
		OpNode o=(OpNode)n;
		Object item1=evaluate(obj,var,o.m_Expr1);
		switch(n.getType()){
		case NodeTypes.s_not:
		    if (boolValue(item1))
			ret=Boolean.FALSE;
		    else
			ret=Boolean.TRUE;
		    break;
		case NodeTypes.s_and:
		    if (boolValue(item1) && boolValue(evaluate(obj,var,o.m_Expr2)))
			ret=Boolean.TRUE;
		    else
			ret=Boolean.FALSE;
		    break;
		case NodeTypes.s_or:
		    if (boolValue(item1) || boolValue(evaluate(obj,var,o.m_Expr2)))
			ret=Boolean.TRUE;
		    else
			ret=Boolean.FALSE;
		    break;
		case NodeTypes.s_Equals:
		    other=evaluate(obj,var,o.m_Expr2);
		    if (other!=null && item1.equals(other))
			ret=Boolean.TRUE;
		    else
			ret=Boolean.FALSE;
		    break;
		case NodeTypes.s_NotEquals:
		    other=evaluate(obj,var,o.m_Expr2);
		    if (other!=null && item1.equals(other))
			ret=Boolean.FALSE;
		    else
			ret=Boolean.TRUE;
		    break;
		case NodeTypes.s_Concat:{
		    String str1="";
		    if (item1!=null)
			str1=item1.toString();
		    other=evaluate(obj,var,o.m_Expr2);
		    String str2="";
		    if (other!=null)
			str2=other.toString();
		    ret=addStrings(str1,evaluate(obj,var,o.m_Expr2).toString());
		    break;
		}
		case NodeTypes.s_Minus:
		    if (item1!=null){
			if (item1 instanceof Double)
			    ret=new Double(-((Double)item1).doubleValue());
			else
			    ret=new Integer(-intValue(item1));
		    }else
			ret=null;
		    break;
		default:{
			Object item2=evaluate(obj,var,o.m_Expr2);
			if (item1!=null && item2!=null){
			    boolean useIntegerArith=(item1 instanceof Integer && item2 instanceof Integer);
			    if (useIntegerArith){
				int val1=intValue(item1);
				int val2=intValue(item2);
				switch(n.getType()){
				case NodeTypes.s_More:
				    if (val1>val2)
					ret=Boolean.TRUE;
				    else
					ret=Boolean.FALSE;
				    break;
				case NodeTypes.s_Less:
				    if (val1<val2)
					ret=Boolean.TRUE;
				    else
					ret=Boolean.FALSE;
				    break;
				case NodeTypes.s_Plus:
				    ret=new Integer(val1+val2);
				    break;
				case NodeTypes.s_Subtract:
				    ret=new Integer(val1-val2);
				    break;
				case NodeTypes.s_Divide:
				    if (val2==0)
					runtimeException("Divide by zero error");
				    else
					ret=new Integer((int)((long)val1/(long)val2));
				    break;
				case NodeTypes.s_Mult:
				    ret=new Integer(val1*val2);break;
				case NodeTypes.s_Mod:
				    ret=new Integer(val1%val2);break;
				}
				if (g_TraceExecution) Tracer.trace("evaluate value1="+val1+" value2="+val2+" result of "+NodeTypes.g_NodeNames[n.getType()]+"="+intValue(ret));
			    }else{
				double val1=floatValue(item1);
				double val2=floatValue(item2);
				switch(n.getType()){
				case NodeTypes.s_More:
				    if (val1>val2)
					ret=Boolean.TRUE;
				    else
					ret=Boolean.FALSE;
				    break;
				case NodeTypes.s_Less:
				    if (val1<val2)
					ret=Boolean.TRUE;
				    else
					ret=Boolean.FALSE;
				    break;
				case NodeTypes.s_Plus:
				    ret=new Double(val1+val2);break;
				case NodeTypes.s_Subtract:
				    ret=new Double(val1-val2);break;
				case NodeTypes.s_Divide:
				    if (val2==0)
					runtimeException("Divide by zero error");
				    else
					ret=new Double(val1/val2);
				    break;
				case NodeTypes.s_Mult:
				    ret=new Double(val1*val2);break;
				case NodeTypes.s_Mod:
				    ret=new Double(val1%val2);break;
				}
				if (g_TraceExecution) Tracer.trace("evaluate value1="+val1+" value2="+val2+" result of "+NodeTypes.g_NodeNames[n.getType()]+"="+floatValue(ret));
			    }
			}else
			    ret=null;
		    }
		}
	    }
	}
	if (g_TraceExecution) Tracer.trace("result of "+NodeTypes.g_NodeNames[n.getType()]+"="+((ret!=null)?ret.toString():"null"));
	return ret;
    }
    /**
     * This method parses a string full of Simkin code and returns 
     * the parse tree. If a parse error occurs, a ParseException is thrown
     * <p>For example the following string could be passed:
     * <p>
     <pre>(a,b){
     a=b+1;
     return a;
     }
     </pre>
     * @param location used to help identify the code in error messages
     * @param code a string containing the Simkin code
     * @return a ParseNode representing the parse tree for the code
     */
    public ParseNode parse(String location,String code) throws ParseException{
	StringBufferInputStream buffer=new StringBufferInputStream(code);
	Parser parser = new Parser(buffer);
	ParseNode topNode=null;
	try{
	    topNode=parser.executable_code();
	}catch(ParseException e){
	    e.location=location;
	    throw e;
	}catch(TokenMgrError e){
	    ParseException pe=new ParseException(e.getMessage());
	    pe.location=location;
	    throw pe;
	}
	return topNode;
    }
    /**
     * This method parses a string of Simkin statements, excluding parameters and enclosing braces.
     * The method returns a parse tree. If a parse error occurs, a ParseException is thrown
     * <p> For example the following string could be passed (note the absence of enclosing braces around the statement list):
     <pre>
     a=b+1;
     return a;
     </pre>
     * @param location used to help identify the code in error messages
     * @param paramNames a vector of names for the parameters
     * @param code a string containing the Simkin code
     * @return a ParseNode representing the parse tree for the code
     */
    public ParseNode parseExternalParams(String location,Vector paramNames,String code) throws ParseException{
	StringBufferInputStream buffer=new StringBufferInputStream(code);
	Parser parser = new Parser(buffer);
	MethodDefNode methodNode=new MethodDefNode();
	for (int i=0;i<paramNames.size();i++)
	    methodNode.m_Params.addId(new IdNode((String)paramNames.elementAt(i),null));
	try{
	    methodNode.m_Stats=parser.stat_list();
	    //	    if (g_TraceExecution) topNode.write(System.out,0);
	}catch(ParseException e){
	    e.location=location;
	    throw e;
	}catch(TokenMgrError e){
	    ParseException pe=new ParseException(e.getMessage());
	    pe.location=location;
	    throw pe;
	}
	return methodNode;
    }
    /**
     * This method executes a piece of Simkin code within a string
     * <p>For example the following string could be passed:
     * <p>
     <pre>(a,b){
     a=b+1;
     return a;
     }
     </pre>
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param code a string containing some Simkin code
     * @param args and array of RValue arguments to the code
     * @return returns an object containing the result of the method and the parse tree for the method, which can be safely cached
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public ExecuteResult executeString(String location,Object obj,String code,Object [] args) throws RuntimeException, ParseException{
	ParseNode parseNode=parse(location,code);
	Object ret=executeParseTree(location,obj,parseNode,args);
	return new ExecuteResult(ret,parseNode);
    }
    /**
     * This method executes a piece of Simkin code within a string, it assumes that the parameters have been defined outside the statement block.
     * <p> For example the following string could be passed (note the absence of enclosing braces around the statement list):
     <pre>
     a=b+1;
     return a;
     </pre>
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param paramNames a vector of names for the parameters
     * @param code a string containing some Simkin code
     * @param args and array of RValue arguments to the code
     * @return returns an object containing the result of the method and the parse tree for the method, which can be safely cached
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public ExecuteResult executeStringExternalParams(String location,Object obj,Vector paramNames,String code,Object[] args) throws RuntimeException, ParseException{
	ParseNode parseNode=parseExternalParams(location,paramNames,code);
	Object ret=executeParseTree(location,obj,parseNode,args);
	return new ExecuteResult(ret,parseNode);
    }
    /**
     * This method executes a piece of Simkin code from a parse tree.<p>You can call this with a parse tree produced by {@link simkin.Interpreter#parse parse}, {@link simkin.Interpreter#parseExternalParams parseExternalParams},{@link simkin.Interpreter#executeString executeString} or {@link simkin.Interpreter#executeStringExternalParams executeStringExternalParams}. This can be useful for caching parse trees to improve performance.
     * @param location a name used to identify the code in error messages
     * @param obj the object within whose context the code will execute
     * @param parseNode a pre-parsed tree of Simkin code
     * @param args and array of RValue arguments to the code
     * @returns the return value of the method call
     * @throws ParseException this is thrown if there is a syntax error in the code
     * @throws RuntimeException this is thrown if there is an error during the execution of the code (such as method or field not found)
     */
    public Object executeParseTree(String location,Object obj,ParseNode parseNode,Object[] args) throws RuntimeException, ParseException{
	String old_location=m_Location;
	m_Location=location;
	Object ret=null;
	Hashtable pvars;
	// if the stack is not too deep, save some time by using our pool of local variable tables
	if (parseNode!=null && parseNode.getType()==NodeTypes.s_MethodDefn){
	    MethodDefNode methodNode=(MethodDefNode)parseNode;
	    if (m_StackDepth<MAX_LOCAL_VARS_CACHE)
		pvars=m_LocalVars[m_StackDepth];
	    else	
		// otherwise create a new table for the local variables.
		pvars=new Hashtable();
	    // fix up parameters by putting them in the local variable table
	    if (args!=null && args.length>0){
		for (int i=0;i<methodNode.m_Params.m_Ids.size();i++){
		    IdNode idNode=(IdNode)(methodNode.m_Params.m_Ids.elementAt(i));
		    String pparam=idNode.m_Name;
		    addLocalVariable(pvars,pparam,args[i]);
		}
	    }
	    m_StackDepth++;
	    ret=executeStats(obj,pvars,methodNode.m_Stats);
	    m_StackDepth--;
	    if (m_StackDepth<MAX_LOCAL_VARS_CACHE){
		pvars.clear();
	    }
	}
	m_Location=old_location;
	return ret;
    }
    /** 
     * this method is called when there is a runtime error and throws a RuntimeException
     */
    static public void runtimeError(String buffer) throws RuntimeException{
	throw new RuntimeException(buffer);
    }
    /** 
     * this method is called when there is a runtime error and throws a RuntimeException, it includes the location in the exception
     */
    void runtimeException(String buffer) throws RuntimeException{
	StringBuffer msg=new StringBuffer(m_Location);
	msg.append(":");
	msg.append(buffer);
	throw new RuntimeException(msg.toString());
    }
    /**
     * this method is called when the interpreter wants to output a message to the tracing log
     */
    void traceString(String buffer){
	Tracer.trace(buffer);
    }
    /*
     * Searches a string for escape sequences \n and \t and converts them
     */
    static String replaceEscapes(String str){
	str=replace(str,"\\n","\n");
	str=replace(str,"\\\"","\"");
	str=replace(str,"\\t","\t");
	return str;
    }
    /**
     * @param original - the string to be replaced
     * @param old_value - the value to be replaced
     * @param new_value - the  new value
     * @return the original string with the old value replaced by the new
     */
    static String replace(String original,String old_value,String new_value){
	String replaced=original;
	int index=original.indexOf(old_value);
	if (index>=0){
	    int pos=0;
	    StringBuffer buffer=new StringBuffer(original.length());
	    while (index>=0){
		buffer.append(original.substring(pos,index));
		buffer.append(new_value);
		pos=index+old_value.length();
		index=original.indexOf(old_value,pos);
	    }
	    if (pos<original.length())
		buffer.append(original.substring(pos));
	    replaced=buffer.toString();
	}
	return replaced;
    }
    /**
     * This method returns the global interpreter associated with all method calls.
     * @return the Interpeter that will be used to execute methods
     */
    public static Interpreter getInterpreter(){
	if (g_Interpreter==null)
	    Interpreter.runtimeError("No global Interpreter defined");
	return g_Interpreter;
    }
    /**
     * Sets a global interpreter to be used for method calls - you can use this to override the Interpreter with your own derived classes
     * @param the Interpeter that will be used to execute methods
     */
    public static void setInterpreter(Interpreter i){
	g_Interpreter=i;
    }
    /**
     * This function attempts to make a public method call on a Java object by using reflection
     * @param owner - the object to call a method on
     * @param method_name - the name of the method to call
     * @param arguments - the method arguments - types Integer, Double, Boolean and Character are converted to their primitive types
     * @return the result of the method call
     * @throws RuntimeException - if the method could not be called
     * @throws NoSuchMethodException - if the method could not be found
     */
    public Object reflectiveMethodCall(Object owner,String method_name,Object[] arguments) throws RuntimeException,NoSuchMethodException{
	Object ret=null;
	if (g_TraceExecution) Tracer.trace("reflectiveMethodCall "+method_name+" on "+owner.getClass().getName());
	Class[] arg_types=null;
	if (arguments!=null && arguments.length>0){
	    arg_types=new Class[arguments.length];
	    for (int i=0;i<arguments.length;i++){
		if (arguments[i] instanceof Integer)
		    arg_types[i]=Integer.TYPE;
		else if (arguments[i] instanceof Double)
		    arg_types[i]=Double.TYPE;
		else if (arguments[i] instanceof Boolean)
		    arg_types[i]=Boolean.TYPE;
		else
		    arg_types[i]=arguments[i].getClass();
	    }
	}
	Method method=null;
	try{
	    method=owner.getClass().getMethod(method_name,arg_types);
	}catch(NoSuchMethodException e){
	    throw e;
	}
	catch(Exception e){
	    runtimeError("Could not find method \""+method_name+"\" is it declared public?");
	}
	if (method!=null){
	    try{
		ret=method.invoke(owner,arguments);
		if (ret==null)
		    ret=g_null;
	    }catch(Exception e){
		runtimeError("Could not call method \""+method_name+"\" it should be declared public");
	    }
	}
	if (g_TraceExecution) Tracer.trace("reflectiveMethodCall "+method_name+" returns "+(ret!=null?ret.toString():"null"));
	return ret;
    }
    /**
     * This function attempts to set a public Java field using reflection
     * @param owner - the object owning the field
     * @param field_name - the name of the field
     * @param value - the value to assign to the field
     * @throws RuntimeException - if the field could not be set (e.g. if it was private)
     * @throws NoSuchFieldException - if the field could not be found
     */
    public void reflectiveSetValue(Object owner,String field_name,Object value) throws RuntimeException,NoSuchFieldException{
	boolean bFound=false;
	Field field=null;
	Class field_class=null;
	try{
	    field=owner.getClass().getField(field_name);
	    field_class=field.getType();
	    bFound=true;	
	}catch(NoSuchFieldException e){
	    throw e;
	}catch(Exception e){
	    runtimeError("Could not access field \""+field_name+"\" it should be declared public");
	}
	if (bFound){
	    try{
		field.set(owner,value);
	    }catch(Exception e){
		runtimeError("Could not set value into field \""+field_name+"\", the field should be declared public");
	    }
	}
    }
    /**
     * This function attempts to retrieve a public Java field using reflection
     * @param owner - the object owning the field
     * @param field_name - the name of the field
     * @return the value of the field
     * @throws RuntimeException - if the field could not be retrieved (e.g. if it was private)
     * @throws NoSuchFieldException - if the field could not be found
     */
    public Object reflectiveGetValue(Object owner,String field_name) throws RuntimeException,NoSuchFieldException{
	Object fieldValue=null;
	Field field=null;
	try{
	    field=owner.getClass().getField(field_name);
	}catch(NoSuchFieldException e){
	    throw e;
	}catch(Exception e){
	    runtimeError("Could not access field \""+field_name+"\" it should be declared public");
	}
	try{
	    fieldValue=field.get(owner);
	}catch(Exception e){
	    runtimeError("Could not retrieve value from field \""+field_name+"\", the field should be declared public");
	}
	if (fieldValue==null)
	    fieldValue=g_null;
	return fieldValue;
    }
    /**
     * Adds two strings together efficiently
     */
    static public String addStrings(String one,String two){
	StringBuffer buffer=new StringBuffer(one.length()+two.length());
	buffer.append(one);
	buffer.append(two);
	return buffer.toString();
    }
    /**
     * Adds three strings together efficiently
     */
    static public String addStrings(String one,String two,String three){
	StringBuffer buffer=new StringBuffer(one.length()+two.length()+three.length());
	buffer.append(one);
	buffer.append(two);
	buffer.append(three);
	return buffer.toString();
    }
    /**
     * Create and assign a default global interpreter
     */
    static {
	g_Interpreter=new Interpreter();
    }
};
